#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   label_map_pb2.py
@Author  :   Wei Loon Cheng
@Version :   1.0
@Contact :   hello@datature.io
@License :   Datature Outpost License 1.0
@Desc    :   Label map protocol buffer definition.

isort:skip_file
"""

# pylint: skip-file

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: label_map.proto
# Protobuf Python Version: 5.29.4
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 5, 29, 4, "", "label_map.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0flabel_map.proto\x12\x19\x64\x61tature.inference_server")\n\rLabelMapEntry\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t"B\n\x08LabelMap\x12\x36\n\x04item\x18\x01 \x03(\x0b\x32(.datature.inference_server.LabelMapEntry'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "label_map_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_LABELMAPENTRY"]._serialized_start = 46
    _globals["_LABELMAPENTRY"]._serialized_end = 87
    _globals["_LABELMAP"]._serialized_start = 89
    _globals["_LABELMAP"]._serialized_end = 155
# @@protoc_insertion_point(module_scope)
